@@ -1,5 +1,5 @@
 /*
- * $Header: /home/t-ishii/repository/pgbench/pgbench.c,v 1.9 1999/09/08 04:27:06 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgbench/pgbench.c,v 1.14 1999/09/10 08:58:14 t-ishii Exp $
  *
  * pgbench: a simple TPC-B like benchmark program for PostgreSQL
  * written by Tatsuo Ishii
@@ -85,7 +85,7 @@ typedef struct {
 } CState;
 
 static void usage() {
-  fprintf(stderr,"usage: pgbench [-h hostname][-p port][-c nclients][-t ntransactions][-s scaling_factor][-n][-d][dbname]\n");
+  fprintf(stderr,"usage: pgbench [-h hostname][-p port][-c nclients][-t ntransactions][-s scaling_factor][-n][-S][-d][dbname]\n");
   fprintf(stderr,"(initialize mode): pgbench -i [-h hostname][-p port][-s scaling_factor][-d][dbname]\n");
 }
 
@@ -109,7 +109,7 @@ static int check(CState *state, PGresult *res, int n, int good)
   CState *st = &state[n];
 
   if (res && PQresultStatus(res) != good) {
-    fprintf(stderr,"client %d:%s",n,PQerrorMessage(st->con));
+    fprintf(stderr,"Client %d aborted in state %d: %s",n,st->state,PQerrorMessage(st->con));
     remains--;	/* I've aborted */
     PQfinish(st->con);
     st->con = NULL;
@@ -129,8 +129,15 @@ static void doOne(CState *state, int n, int debug) {
       fprintf(stderr,"client %d receiving\n",n);
     }
     while (PQisBusy(st->con) == TRUE) {
-      PQconsumeInput(st->con);
+      if (!PQconsumeInput(st->con)) {	/* there's something wrong */
+	fprintf(stderr, "Client %d aborted in state %d. Probably the backend died while processing.\n",n, st->state);
+	remains--;	/* I've aborted */
+	PQfinish(st->con);
+	st->con = NULL;
+	return;
+      }
     }
+
     switch (st->state) {
     case 0:	/* response to "begin" */
       res = PQgetResult(st->con);
@@ -247,6 +254,72 @@ static void doOne(CState *state, int n, int debug) {
   }
 }
 
+/* process a select only transaction */
+static void doSelectOnly(CState *state, int n, int debug) {
+  char sql[256];
+  PGresult   *res;
+  CState *st = &state[n];
+
+  if (st->listen) {	/* are we receiver? */
+    if (debug) {
+      fprintf(stderr,"client %d receiving\n",n);
+    }
+    while (PQisBusy(st->con) == TRUE) {
+      if (!PQconsumeInput(st->con)) {	/* there's something wrong */
+	fprintf(stderr, "Client %d aborted in state %d. Probably the backend died while processing.\n",n, st->state);
+	remains--;	/* I've aborted */
+	PQfinish(st->con);
+	st->con = NULL;
+	return;
+      }
+    }
+
+    switch (st->state) {
+    case 0:	/* response to "select abalance ..." */
+      res = PQgetResult(st->con);
+      if (check(state, res, n, PGRES_TUPLES_OK)) {
+	return;
+      }
+      PQclear(res);
+      discard_response(st);
+
+      if (++st->cnt >= nxacts) {
+	remains--;	/* I've done */
+	PQfinish(st->con);
+	st->con = NULL;
+	return;
+      }
+      break;
+    }
+
+    /* increment state counter */
+    st->state++;
+    if (st->state > 0) {
+      st->state = 0;
+    }
+  }
+
+  switch (st->state) {
+  case 0:
+    st->aid = getrand(1,naccounts*tps);
+    sprintf(sql,"select abalance from accounts where aid = %d",st->aid);
+    break;
+  }
+
+  if (debug) {
+    fprintf(stderr,"client %d sending %s\n",n,sql);
+  }
+
+  if (PQsendQuery(st->con, sql) == 0) {
+    if (debug) {
+      fprintf(stderr, "PQsendQuery(%s)failed\n",sql);
+    }
+    st->ecnt++;
+  } else {
+    st->listen++;	/* flags that should be listned */
+  }
+}
+
 /* discard connections */
 static void disconnect_all(CState *state) {
   int i;
@@ -366,7 +439,7 @@ void init(char *pghost, char *pgport,char *dbName) {
 }
 
 /* print out results */
-printResults(CState *state, struct timeval *tv1,struct timeval *tv2,struct timeval *tv3) {
+printResults(int ttype, CState *state, struct timeval *tv1,struct timeval *tv2,struct timeval *tv3) {
   double t1,t2;
   int i;
   int error_xacts = 0;
@@ -376,12 +449,13 @@ printResults(CState *state, struct timeval *tv1,struct timeval *tv2,struct timev
     normal_xacts += state[i].cnt;
   }
 
-  t1 = (tv3->tv_sec - tv1->tv_sec)*1000000+(tv3->tv_usec - tv1->tv_usec);
+  t1 = (tv3->tv_sec - tv1->tv_sec)*1000000.0+(tv3->tv_usec - tv1->tv_usec);
   t1 = normal_xacts*1000000.0/t1;
 
-  t2 = (tv3->tv_sec - tv2->tv_sec)*1000000+(tv3->tv_usec - tv2->tv_usec);
+  t2 = (tv3->tv_sec - tv2->tv_sec)*1000000.0+(tv3->tv_usec - tv2->tv_usec);
   t2 = normal_xacts*1000000.0/t2;
 
+  printf("transaction type: %s\n",ttype==0?"TPC-B (sort of)":"SELECT only");
   printf("scaling factor: %d\n",tps);
   printf("number of clients: %d\n",nclients);
   printf("number of transactions per client: %d\n",nxacts);
@@ -400,6 +474,7 @@ int main(int argc, char **argv) {
   int is_init_mode = 0;	/* initialize mode? */
   int is_vacuum = 1;	/* do vacuum before testing? */
   int debug = 0;	/* debug flag */
+  int ttype = 0;	/* transaction type. 0: TPC-B, 1: SELECT only */
 
   static CState	   state[MAXCLIENTS];	/* clients status */
 
@@ -411,16 +486,17 @@ int main(int argc, char **argv) {
   char ebuf[128];	/* error string buffer */
 
   fd_set	input_mask;
-  fd_set	output_mask;
-  static struct timeval timeout = {10000,0};
   int nsocks;	/* return from select(2) */
   int maxsock;	/* max socket number to be waited */
+
+#ifndef __CYGWIN32__
   struct rlimit rlim;
+#endif
 
   PGconn *con;
   PGresult *res;
 
-  while ((c = getopt(argc, argv, "ih:np:dc:t:s:")) != EOF)	{
+  while ((c = getopt(argc, argv, "ih:np:dc:t:s:S")) != EOF) {
     switch (c) {
     case 'i':
       is_init_mode++;
@@ -437,19 +513,22 @@ int main(int argc, char **argv) {
     case 'd':
       debug++;
       break;
+    case 'S':
+      ttype = 1;
+      break;
     case 'c':
       nclients = atoi(optarg);
       if (nclients <= 0 || nclients > MAXCLIENTS) {
 	fprintf(stderr,"wrong number of clients: %d\n",nclients);
 	exit(1);
       }
+#ifndef __CYGWIN32__
 #ifdef HAVE_RLIMIT_NOFILE	/* most platform uses RLIMIT_NOFILE */
       if (getrlimit(RLIMIT_NOFILE,&rlim) == -1) {
 #else	/* but BSD doesn't ... */
       if (getrlimit(RLIMIT_OFILE,&rlim) == -1) {
 #endif /* HAVE_RLIMIT_NOFILE */
-	perror(ebuf);
-	fprintf(stderr,"getrlimit failed. reason: %s\n",ebuf);
+	fprintf(stderr,"getrlimit failed. reason: %s\n",strerror(errno));
 	exit(1);
       }
       if (rlim.rlim_cur <= (nclients+2)) {
@@ -457,6 +536,7 @@ int main(int argc, char **argv) {
 	fprintf(stderr,"Use limit/ulimt to increase the limit before using pgbench.\n");
 	exit(1);
       }
+#endif /* #ifndef __CYGWIN32__ */
       break;
     case 's':
       tps = atoi(optarg);
@@ -508,7 +588,7 @@ int main(int argc, char **argv) {
     exit(1);
   }
 
-  /* get scaling factor that should be same as count(*) from branches... */
+  /* get the scaling factor that should be same as count(*) from branches... */
   res = PQexec(con,"select count(*) from branches");
   if (PQresultStatus(res) != PGRES_TUPLES_OK) {
     fprintf(stderr, "%s", PQerrorMessage(con));
@@ -563,17 +643,19 @@ int main(int argc, char **argv) {
 
   /* send start up quries in async manner */
   for (i=0;i<nclients;i++) {
-    doOne(state, i, debug);
+    if (ttype == 0) {
+      doOne(state, i, debug);
+    } else if (ttype == 1) {
+      doSelectOnly(state, i, debug);
+    }
   }
 
-  FD_ZERO(&output_mask);
-
   for (;;) {
-    if (remains == 0) {	/* all done ? */
+    if (remains <= 0) {	/* all done ? */
       disconnect_all(state);
       /* get end time */
       gettimeofday(&tv3, 0);
-      printResults(state, &tv1,&tv2,&tv3);
+      printResults(ttype, state, &tv1,&tv2,&tv3);
       exit(0);
     }
 
@@ -595,15 +677,14 @@ int main(int argc, char **argv) {
       }
     }
 
-    if ((nsocks = select(maxsock +1, &input_mask, &output_mask,
-			 (fd_set *) NULL, &timeout)) < 0) {
+    if ((nsocks = select(maxsock +1, &input_mask, (fd_set *)NULL,
+			 (fd_set *)NULL, (struct timeval *)NULL)) < 0) {
       if (errno == EINTR) {
 	continue;
       }
       /* must be something wrong */
       disconnect_all(state);
-      perror(ebuf);
-      fprintf(stderr,"select failed: %s\n",ebuf);
+      fprintf(stderr,"select failed: %s\n",strerror(errno));
       exit(1);
     } else if (nsocks == 0) { /* timeout */
       fprintf(stderr,"select timeout\n");
@@ -617,7 +698,11 @@ int main(int argc, char **argv) {
     /* ok, backend returns reply */
     for (i=0;i<nclients;i++) {
       if (state[i].con && FD_ISSET(PQsocket(state[i].con), &input_mask)) {
-	doOne(state, i, debug);
+	if (ttype == 0) {
+	  doOne(state, i, debug);
+	} else if (ttype == 1) {
+	  doSelectOnly(state, i, debug);
+	}
       }
     }
   }
